"""
The discovery based on :term:`Consul`.

This driver implement a client side service discovery.
"""
import random
from typing import Any, Callable, List

from pydantic.fields import Field

from blacksmith.domain.exceptions import HTTPError, UnregisteredServiceException
from blacksmith.domain.model import PathInfoField, Request, Response
from blacksmith.domain.model.params import CollectionIterator
from blacksmith.domain.registry import Registry
from blacksmith.middleware._async.auth import AsyncHTTPBearerAuthorization
from blacksmith.sd._async.adapters.static import AsyncStaticDiscovery
from blacksmith.sd._async.base import AsyncAbstractServiceDiscovery, Url
from blacksmith.service._async.client import AsyncClientFactory
from blacksmith.typing import ServiceName, Version


class ConsulApiError(HTTPError):
    """Raised when consul API is not responding what is expected."""

    def __init__(self, exc: HTTPError):
        return super().__init__(str(exc), exc.request, exc.response)


class ServiceRequest(Request):
    """Request parameter of the Consul API to retrieve a host for a service."""

    name: str = PathInfoField()
    """Name of the service to search for an endpoint."""


class Service(Response):
    """Consul Service response."""

    address: str = Field(alias="ServiceAddress")
    """IP Address of an instance that host the service."""
    port: int = Field(alias="ServicePort")
    """TCP Port of an instance that host the service."""


_registry = Registry()
_registry.register(
    "consul",
    "services",
    "consul",
    "v1",
    collection_path="/catalog/service/{name}",
    collection_contract={"GET": (ServiceRequest, Service)},
)


def blacksmith_cli(
    endpoint: Url, consul_token: str
) -> AsyncClientFactory[Service, Any]:
    sd = AsyncStaticDiscovery({("consul", "v1"): endpoint})
    fact: AsyncClientFactory[Service, Any] = AsyncClientFactory(sd, registry=_registry)
    if consul_token:
        fact.add_middleware(AsyncHTTPBearerAuthorization(consul_token))
    return fact


class AsyncConsulDiscovery(AsyncAbstractServiceDiscovery):
    """
    A discovery instance based on a :term:`Consul` server.

    :param service_name_fmt: pattern for name of versionned service
    :param service_url_fmt: pattern for url of versionned service
    :param unversioned_service_name_fmt: pattern for name of unversioned service
    :param unversioned_service_url_fmt: pattern for url of unversioned service

    """

    service_name_fmt: str
    service_url_fmt: str
    unversioned_service_name_fmt: str
    unversioned_service_url_fmt: str

    def __init__(
        self,
        addr: Url = "http://consul:8500/v1",
        service_name_fmt: str = "{service}-{version}",
        service_url_fmt: str = "http://{address}:{port}/{version}",
        unversioned_service_name_fmt: str = "{service}",
        unversioned_service_url_fmt: str = "http://{address}:{port}",
        consul_token: str = "",
        _client_factory: Callable[
            [Url, str], AsyncClientFactory[Service, Any]
        ] = blacksmith_cli,
    ) -> None:
        self.blacksmith_cli = _client_factory(addr, consul_token)
        self.service_name_fmt = service_name_fmt
        self.service_url_fmt = service_url_fmt
        self.unversioned_service_name_fmt = unversioned_service_name_fmt
        self.unversioned_service_url_fmt = unversioned_service_url_fmt

    def format_service_name(self, service: ServiceName, version: Version) -> str:
        """Build the service name to send to consul."""
        if version is None:
            name = self.unversioned_service_name_fmt.format(service=service)
        else:
            name = self.service_name_fmt.format(service=service, version=version)
        return name

    def format_endoint(self, version: Version, address: str, port: int) -> Url:
        """Build the rest api endpoint from consul response."""
        if version is None:
            endpoint = self.unversioned_service_url_fmt.format(
                address=address, port=port
            )
        else:
            endpoint = self.service_url_fmt.format(
                version=version, address=address, port=port
            )
        return endpoint

    async def resolve(self, service: ServiceName, version: Version) -> Service:
        """
        Get the :class:`Service` from the consul registry.

        If many instances host the service, the host is choosen randomly.
        """
        name = self.format_service_name(service, version)
        consul = await self.blacksmith_cli("consul")
        try:
            iresp: CollectionIterator[Service] = await consul.services.collection_get(
                ServiceRequest(name=name)
            )
        except HTTPError as exc:
            raise ConsulApiError(exc)  # rewrite the class to avoid confusion
        else:
            resp: List[Service] = list(iresp)
            if not resp:
                raise UnregisteredServiceException(service, version)
            return random.choice(resp)

    async def get_endpoint(self, service: ServiceName, version: Version) -> Url:
        """
        Get the endpoint from the consul registry

        If many instances host the service, the host is choosen randomly.
        """
        srv = await self.resolve(service, version)
        return self.format_endoint(version, srv.address, srv.port)



from typing import Dict

from blacksmith.domain.model.http import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.domain.typing import AsyncMiddleware
from blacksmith.typing import ClientName, Path


class AsyncHTTPMiddleware:
    """Inject data in http query on every requests."""

    def __init__(self) -> None:
        pass

    async def initialize(self) -> None:
        pass

    def __call__(self, next: AsyncMiddleware) -> AsyncMiddleware:
        async def handle(
            req: HTTPRequest,
            client_name: ClientName,
            path: Path,
            timeout: HTTPTimeout,
        ) -> HTTPResponse:
            return await next(req, client_name, path, timeout)

        return handle


class AsyncHTTPAddHeadersMiddleware(AsyncHTTPMiddleware):
    """
    Generic middleware that inject HTTP headers.

    :params: headers to inject in HTTP requests.
    """

    headers: Dict[str, str]

    def __init__(self, headers: Dict[str, str]):
        self.headers = headers

    def __call__(self, next: AsyncMiddleware) -> AsyncMiddleware:
        async def handle(
            req: HTTPRequest,
            client_name: ClientName,
            path: Path,
            timeout: HTTPTimeout,
        ) -> HTTPResponse:
            req.headers.update(self.headers)
            return await next(req, client_name, path, timeout)

        return handle



import abc

from blacksmith.typing import ServiceName, Url, Version


class AsyncAbstractServiceDiscovery(abc.ABC):
    """Define the Service Discovery interface."""

    @abc.abstractmethod
    async def get_endpoint(self, service: ServiceName, version: Version) -> Url:
        """Get the endpoint of a service."""



import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional, Union, cast

from blacksmith.typing import HTTPMethod, Json, Url

simpletypes = Union[str, int, float, bool]
Links = Dict[Optional[str], Dict[str, str]]


class HTTPTimeout:
    """Request timeout."""

    request: float
    connect: float

    def __init__(self, timeout: float = 30.0, connect: float = 15.0) -> None:
        self.request = timeout
        self.connect = connect

    def __eq__(self, other: Any) -> bool:
        return self.request == other.request and self.connect == other.connect


@dataclass
class HTTPRequest:
    """
    Internal representation of an http request.

    Note that the HTTP method is not present, because the method is
    the funcion called.

    The HTTP Request is filled out using the
    :class:`blacksmith.domain.model.params.Request` schema.
    """

    method: HTTPMethod
    url_pattern: Url
    # the property match with the "location" of feaut
    path: Dict[str, simpletypes] = field(default_factory=dict)
    querystring: Dict[str, Union[simpletypes, List[simpletypes]]] = field(
        default_factory=dict
    )
    headers: Dict[str, str] = field(default_factory=dict)
    body: str = ""

    @property
    def url(self) -> str:
        return self.url_pattern.format(**self.path)


def parse_header_links(value: str) -> List[Dict[str, str]]:
    """
    Returns a list of parsed link headers, for more info see:
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link

    The generic syntax of those is:

    ::

        Link: < uri-reference >; param1=value1; param2="value2"

    So for instance:

    Link; '<http:/.../front.jpeg>; type="image/jpeg",<http://.../back.jpeg>;'
    would return

    ::

        [
            {"url": "http:/.../front.jpeg", "type": "image/jpeg"},
            {"url": "http://.../back.jpeg"},
        ]

    .. note::

        Stolen code from httpx _utils.py (private method)


    :param value: HTTP Link entity-header field
    :return: list of parsed link headers
    """
    links: List[Dict[str, str]] = []
    replace_chars = " '\""
    value = value.strip(replace_chars)
    if not value:
        return links
    for val in re.split(", *<", value):
        try:
            url, params = val.split(";", 1)
        except ValueError:
            url, params = val, ""
        link = {"url": url.strip("<> '\"")}
        for param in params.split(";"):
            try:
                key, value = param.split("=")
            except ValueError:
                break
            link[key.strip(replace_chars)] = value.strip(replace_chars)
        links.append(link)
    return links


@dataclass
class HTTPResponse:
    """
    Internal representation of an http response.
    """

    status_code: int
    """HTTP Status code."""
    headers: Mapping[str, str]
    """Header of the response."""
    json: Json
    """Json Body of the response."""

    @property
    def links(self) -> Links:
        header = self.headers.get("link")
        ldict = {}
        if header:
            links = parse_header_links(header)
            for link in links:
                key = link.get("rel") or link.get("url")
                ldict[key] = link
        return cast(Links, ldict)



"""Trace with zipkin of jaegger."""

import abc
from typing import Any, Dict, Optional


class AbtractTraceContext(abc.ABC):
    """
    Interface of the trace context for the middleware.

    See examples with starlette-zipking for an implementation.
    """

    @classmethod
    @abc.abstractmethod
    def make_headers(cls) -> Dict[str, str]:
        """Build headers for the sub requests."""

    @abc.abstractmethod
    def __init__(self, name: str, kind: str = "SERVER") -> None:
        """Create a trace span for the current context."""

    @abc.abstractmethod
    def tag(self, key: str, value: str) -> "AbtractTraceContext":
        """Tag the span"""

    @abc.abstractmethod
    def annotate(
        self, value: Optional[str], ts: Optional[float] = None
    ) -> "AbtractTraceContext":
        """Annotate the span"""

    @abc.abstractmethod
    def __enter__(self) -> "AbtractTraceContext":
        """Make the created trace span of the current context the active span."""

    @abc.abstractmethod
    def __exit__(self, *exc: Any) -> None:
        """
        Ends the created trace span of the context, it parents become the active span.
        """



"""Collect metrics based on prometheus."""
import abc
import json
from dataclasses import asdict
from datetime import timedelta
from typing import Any, List, Optional, Tuple, Type
from urllib.parse import urlencode

from httpx import Headers

from blacksmith.domain.model.http import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.typing import ClientName, Path

from .base import SyncHTTPMiddleware, SyncMiddleware


class AbstractCachingPolicy(abc.ABC):
    """Caching Policy"""

    @abc.abstractmethod
    def handle_request(
        self, req: HTTPRequest, client_name: ClientName, path: Path
    ) -> bool:
        """A function to decide if the http request is cachable."""

    @abc.abstractmethod
    def get_vary_key(
        self,
        client_name: ClientName,
        path: Path,
        request: HTTPRequest,
    ) -> str:
        """Create a caching key for the vary part."""

    @abc.abstractmethod
    def get_response_cache_key(
        self,
        client_name: ClientName,
        path: Path,
        req: HTTPRequest,
        vary: List[str],
    ) -> str:
        """Create a caching key for the http response."""

    @abc.abstractmethod
    def get_cache_info_for_response(
        self,
        client_name: ClientName,
        path: Path,
        req: HTTPRequest,
        resp: HTTPResponse,
    ) -> Tuple[int, str, List[str]]:
        """Return caching info. Tuple (ttl in seconds, vary key, vary list)."""


class SyncAbstractCache(abc.ABC):
    """Abstract Redis Client."""

    @abc.abstractmethod
    def initialize(self) -> None:
        """Initialize the cache"""

    @abc.abstractmethod
    def get(self, key: str) -> Optional[str]:
        """Get a value from redis"""

    @abc.abstractmethod
    def set(self, key: str, val: str, ex: timedelta) -> None:
        """Get a value from redis"""


try:
    from redis import Redis

    SyncAbstractCache.register(Redis)
except ImportError:
    pass


class AbstractSerializer(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def loads(s: str) -> Any:
        """Load a string to an object"""

    @staticmethod
    @abc.abstractmethod
    def dumps(obj: Any) -> str:
        """Get a value from redis"""


class JsonSerializer(AbstractSerializer):
    @staticmethod
    def loads(s: str) -> Any:
        return json.loads(s)

    @staticmethod
    def dumps(obj: Any) -> str:
        return json.dumps(obj)


def int_or_0(val: str) -> int:
    try:
        ival = int(val)
    except ValueError:
        ival = 0
    return ival


def get_max_age(response: HTTPResponse) -> int:
    max_age = 0
    age = int_or_0(response.headers.get("age", "0"))
    s_cache_control = response.headers.get("cache-control", "")
    cache_control = [c.strip() for c in s_cache_control.split(",")]
    if "public" in cache_control:
        h_max_age: List[str] = [cc for cc in cache_control if cc.startswith("max-age=")]
        if h_max_age:
            _hdr, value = h_max_age[0].split("=", 1)
            max_age = int_or_0(value)
    return max(max_age - age, 0)


def get_vary_header_split(response: HTTPResponse) -> List[str]:
    vary = response.headers.get("vary", "")
    fields = [field.strip().lower() for field in vary.split(",")] if vary else []
    return fields


class CacheControlPolicy(AbstractCachingPolicy):
    """
    Initialize the caching using `Cache-Control` http headers.
    Also consume the `Vary` response header to cache response per
    Vary response headers per request.

    :param sep: Separator used in cache key **MUST NOT BE USED** in client name.
    """

    def __init__(self, sep: str = "$") -> None:
        self.sep = sep

    def handle_request(
        self, req: HTTPRequest, client_name: ClientName, path: Path
    ) -> bool:
        return req.method == "GET"

    def get_vary_key(
        self, client_name: ClientName, path: Path, request: HTTPRequest
    ) -> str:
        path = path.format(**request.path)
        if request.querystring:
            qs = urlencode(request.querystring, doseq=True)
            path = f"{path}?{qs}"
        return f"{client_name}{self.sep}{path}"

    def get_response_cache_key(
        self,
        client_name: ClientName,
        path: Path,
        req: HTTPRequest,
        vary: List[str],
    ) -> str:
        headers = Headers(req.headers)
        vary_key = self.get_vary_key(client_name, path, req)
        vary_vals = [f"{key}={headers.get(key, '')}" for key in vary]
        response_cache_key = f"{vary_key}{self.sep}{'|'.join(vary_vals)}"
        return response_cache_key

    def get_cache_info_for_response(
        self,
        client_name: ClientName,
        path: Path,
        req: HTTPRequest,
        resp: HTTPResponse,
    ) -> Tuple[int, str, List[str]]:
        max_age = get_max_age(resp)
        if max_age <= 0:
            return (max_age, "", [])
        vary_key = self.get_vary_key(client_name, path, req)
        vary = get_vary_header_split(resp)
        return (max_age, vary_key, vary)


class SyncHTTPCachingMiddleware(SyncHTTPMiddleware):
    """
    Http Cache Middleware based on Cache-Control and redis.
    """

    def __init__(
        self,
        cache: SyncAbstractCache,
        policy: AbstractCachingPolicy = CacheControlPolicy(),
        serializer: Type[AbstractSerializer] = JsonSerializer,
    ) -> None:
        self._cache = cache
        self._policy = policy
        self._serializer = serializer

    def initialize(self) -> None:
        try:
            self._cache.initialize()
        except AttributeError:  # coverage-ignore
            # the redis sync version does not implement this method
            pass

    def cache_response(
        self,
        client_name: ClientName,
        path: Path,
        req: HTTPRequest,
        resp: HTTPResponse,
    ) -> None:
        (
            ttl,
            vary_key,
            vary,
        ) = self._policy.get_cache_info_for_response(client_name, path, req, resp)
        if ttl <= 0:
            return
        ttld = timedelta(seconds=ttl)
        vary_val = self._serializer.dumps(vary)
        self._cache.set(vary_key, vary_val, ttld)

        response_cache_key = self._policy.get_response_cache_key(
            client_name, path, req, vary
        )
        resp.headers = dict(resp.headers)
        response_cache = self._serializer.dumps(asdict(resp))
        self._cache.set(response_cache_key, response_cache, ttld)

    def get_from_cache(
        self, client_name: ClientName, path: Path, req: HTTPRequest
    ) -> Optional[HTTPResponse]:
        vary_key = self._policy.get_vary_key(client_name, path, req)
        vary_val = self._cache.get(vary_key)
        if not vary_val:
            return None
        vary = self._serializer.loads(vary_val)
        response_cache_key = self._policy.get_response_cache_key(
            client_name, path, req, vary
        )
        val = self._cache.get(response_cache_key)
        if not val:
            return None
        resp = self._serializer.loads(val)
        return HTTPResponse(**resp)

    def __call__(self, next: SyncMiddleware) -> SyncMiddleware:
        def handle(
            req: HTTPRequest,
            client_name: ClientName,
            path: Path,
            timeout: HTTPTimeout,
        ) -> HTTPResponse:

            if not self._policy.handle_request(req, client_name, path):
                return next(req, client_name, path, timeout)

            resp = self.get_from_cache(client_name, path, req)
            if resp:
                return resp
            resp = next(req, client_name, path, timeout)
            self.cache_response(client_name, path, req, resp)
            return resp

        return handle



from typing import Optional

from blacksmith.domain.typing import AsyncMiddleware
from blacksmith.typing import Proxies


class AsyncAbstractTransport(AsyncMiddleware):
    def __init__(
        self, verify_certificate: bool = True, proxies: Optional[Proxies] = None
    ):
        self.verify_certificate = verify_certificate
        self.proxies = proxies



"""Trace with zipkin of jaegger."""

import abc
from typing import Any, Dict, Optional, Type

from blacksmith.domain.exceptions import HTTPError
from blacksmith.domain.model.http import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.typing import ClientName, Path

from .base import SyncHTTPMiddleware, SyncMiddleware


class AbtractTraceContext(abc.ABC):
    """
    Interface of the trace context for the middleware.

    See examples with starlette-zipking for an implementation.
    """

    @classmethod
    @abc.abstractmethod
    def make_headers(cls) -> Dict[str, str]:
        """Build headers for the sub requests."""

    @abc.abstractmethod
    def __init__(self, name: str, kind: str = "SERVER") -> None:
        """Create a trace span for the current context."""

    @abc.abstractmethod
    def tag(self, key: str, value: str) -> "AbtractTraceContext":
        """Tag the span"""

    @abc.abstractmethod
    def annotate(
        self, value: Optional[str], ts: Optional[float] = None
    ) -> "AbtractTraceContext":
        """Annotate the span"""

    @abc.abstractmethod
    def __enter__(self) -> "AbtractTraceContext":
        """Make the created trace span of the current context the active span."""

    @abc.abstractmethod
    def __exit__(self, *exc: Any) -> None:
        """
        Ends the created trace span of the context, it parents become the active span.
        """


class SyncZipkinMiddleware(SyncHTTPMiddleware):
    """
    Zipkin Middleware based on an abstract context manager.

    :param trace: A deferred context manager that manage the trace span stack.
    """

    def __init__(self, trace: Type[AbtractTraceContext]) -> None:
        self.trace = trace

    def __call__(self, next: SyncMiddleware) -> SyncMiddleware:
        def handle(
            req: HTTPRequest,
            client_name: ClientName,
            path: Path,
            timeout: HTTPTimeout,
        ) -> HTTPResponse:

            name = f"{req.method} {path.format(**req.path)}"

            with self.trace(name, "CLIENT") as child_span:
                headers = self.trace.make_headers()
                req.headers.update(headers)

                child_span.tag("blacksmith.client_name", client_name)

                child_span.tag("http.path", path)
                if req.querystring:
                    child_span.tag("http.querystring", repr(req.querystring))
                try:
                    resp = next(req, client_name, path, timeout)
                except HTTPError as exc:
                    child_span.tag("http.status_code", str(exc.response.status_code))
                    child_span.tag("error", "true")
                    raise
                else:
                    child_span.tag("http.status_code", str(resp.status_code))
                    return resp

        return handle



from .auth import AsyncHTTPAuthorization, AsyncHTTPBearerAuthorization
from .base import AsyncHTTPAddHeadersMiddleware, AsyncHTTPMiddleware, AsyncMiddleware
from .circuit_breaker import AsyncCircuitBreaker
from .http_caching import AsyncHTTPCachingMiddleware
from .prometheus import AsyncPrometheusMetrics
from .zipkin import AsyncZipkinMiddleware

__all__ = [
    "AsyncCircuitBreaker",
    "AsyncHTTPAddHeadersMiddleware",
    "AsyncHTTPAuthorization",
    "AsyncHTTPBearerAuthorization",
    "AsyncHTTPCachingMiddleware",
    "AsyncHTTPMiddleware",
    "AsyncMiddleware",
    "AsyncPrometheusMetrics",
    "AsyncZipkinMiddleware",
]



"""Register resources on services."""


from collections import defaultdict
from dataclasses import dataclass
from typing import Mapping, MutableMapping, Optional, Tuple, Type

from blacksmith.typing import (
    ClientName,
    HTTPMethod,
    Path,
    ResourceName,
    Service,
    ServiceName,
    Version,
)

from .exceptions import ConfigurationError, UnregisteredClientException
from .model import AbstractCollectionParser, Request, Response

Schemas = Tuple[Type[Request], Optional[Type[Response]]]
Contract = Mapping[HTTPMethod, Schemas]


@dataclass(frozen=True)
class HttpResource:
    """Represent a resource endpoint."""

    path: Path
    """Path that identify the resource."""
    contract: Optional[Contract]
    """A contract is a serialization schema for the request and there response."""


@dataclass(frozen=True)
class HttpCollection(HttpResource):
    collection_parser: Optional[Type[AbstractCollectionParser]]
    """Override the default collection parlser for a given resource."""


class ApiRoutes:
    """
    Store different routes for a type of resource.

    Api may have a route for the resource and/or a route for collection.
    They both have distinct contract for every http method.
    """

    resource: Optional[HttpResource]
    """Resource endpoint"""
    collection: Optional[HttpCollection]
    """Collection endpoint."""

    def __init__(
        self,
        path: Optional[Path],
        contract: Optional[Contract],
        collection_path: Optional[Path],
        collection_contract: Optional[Contract],
        collection_parser: Optional[Type[AbstractCollectionParser]],
    ) -> None:
        self.resource = HttpResource(path, contract) if path else None
        self.collection = (
            HttpCollection(collection_path, collection_contract, collection_parser)
            if collection_path
            else None
        )


Resources = Mapping[ResourceName, ApiRoutes]


class Registry:
    """Store resources in a registry."""

    clients: MutableMapping[ClientName, MutableMapping[ResourceName, ApiRoutes]]
    client_service: MutableMapping[ClientName, Service]

    def __init__(self) -> None:
        self.clients = defaultdict(dict)
        self.client_service = {}

    def register(
        self,
        client_name: ClientName,
        resource: ResourceName,
        service: ServiceName,
        version: Version,
        path: Optional[Path] = None,
        contract: Optional[Contract] = None,
        collection_path: Optional[Path] = None,
        collection_contract: Optional[Contract] = None,
        collection_parser: Optional[Type[AbstractCollectionParser]] = None,
    ) -> None:
        """
        Register the resource in the registry.

        :param client_name: used to identify the client in your code.
        :param resource: name of the resource in your code.
        :param service: name of the service in the service discovery.
        :param version: version number of the service.
        :param path: endpoint of the resource in the given service.
        :param contract: contract for the resource, define request and response.

        :param collection_path: endpoint of the resource collection
            in the given service.
        :param collection_contract: contract for the resource collection,
            define request and response.
        """
        if client_name in self.client_service and self.client_service[client_name] != (
            service,
            version,
        ):
            raise ConfigurationError(
                client_name, self.client_service[client_name], (service, version)
            )

        self.client_service[client_name] = (service, version)
        self.clients[client_name][resource] = ApiRoutes(
            path,
            contract,
            collection_path,
            collection_contract,
            collection_parser,
        )

    def get_service(self, client_name: ClientName) -> Tuple[Service, Resources]:
        """
        Get the service associated for the client.

        This method is used to find the endpoint of the service.
        """
        try:
            return self.client_service[client_name], self.clients[client_name]
        except KeyError:
            raise UnregisteredClientException(client_name)


registry = Registry()
"""Detault registry."""


def register(
    client_name: ClientName,
    resource: ResourceName,
    service: ServiceName,
    version: Version,
    path: Optional[Path] = None,
    contract: Optional[Contract] = None,
    collection_path: Optional[Path] = None,
    collection_contract: Optional[Contract] = None,
    collection_parser: Optional[Type[AbstractCollectionParser]] = None,
) -> None:
    """
    Register a resource in a client in the default registry.

    See :func:`blacksmith.domain.registry.Registry.register` for the signature.
    """
    registry.register(
        client_name,
        resource,
        service,
        version,
        path,
        contract,
        collection_path,
        collection_contract,
        collection_parser,
    )



"""Cut the circuit in case a service is down."""

from typing import Any, Iterable, Optional

from purgatory import AsyncAbstractUnitOfWork, AsyncCircuitBreakerFactory
from purgatory.typing import TTL, Hook, Threshold

from blacksmith.domain.exceptions import HTTPError
from blacksmith.domain.model.http import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.typing import ClientName, Path

from .base import AsyncHTTPMiddleware, AsyncMiddleware
from .prometheus import AsyncPrometheusMetrics

Listeners = Optional[Iterable[Hook]]


def exclude_httpx_4xx(exc: HTTPError) -> bool:
    """Exclude client side http errors."""
    return exc.is_client_error


class GaugeStateValue:
    CLOSED = 0
    HALF_OPEN = 1
    OPEN = 2


class PrometheusHook:
    def __init__(self, prometheus_metrics: AsyncPrometheusMetrics):
        self.prometheus_metrics = prometheus_metrics

    def __call__(self, circuit_name: str, evt_type: str, payload: Any) -> None:
        if evt_type == "state_changed":
            state = {
                "closed": GaugeStateValue.CLOSED,
                "half-opened": GaugeStateValue.HALF_OPEN,
                "opened": GaugeStateValue.OPEN,
            }[payload.state]
            metric = self.prometheus_metrics.blacksmith_circuit_breaker_state
            metric.labels(circuit_name).set(state)
        elif evt_type == "failed":
            metric = self.prometheus_metrics.blacksmith_circuit_breaker_error
            metric.labels(circuit_name).inc()


class AsyncCircuitBreaker(AsyncHTTPMiddleware):
    """
    Prevent cascading failure.

    The circuit breaker is based on `purgatory`_, the middleware create
    one circuit breaker per client_name. The parameters are forwarded
    to all the clients. This middleware does not give the possibility to
    adapt a threshold or the time the circuit is opened per clients.

    .. _`purgatory`: https://pypi.org/project/purgatory-circuitbreaker/
    """

    def __init__(
        self,
        threshold: Threshold = 5,
        ttl: TTL = 30,
        listeners: Listeners = None,
        uow: Optional[AsyncAbstractUnitOfWork] = None,
        prometheus_metrics: Optional[AsyncPrometheusMetrics] = None,
    ):
        self.circuit_breaker = AsyncCircuitBreakerFactory(
            default_threshold=threshold,
            default_ttl=ttl,
            exclude=[(HTTPError, exclude_httpx_4xx)],
            uow=uow,
        )
        if prometheus_metrics:
            self.circuit_breaker.add_listener(PrometheusHook(prometheus_metrics))
        if listeners:
            for listener in listeners:
                self.circuit_breaker.add_listener(listener)

    async def initialize(self) -> None:
        await self.circuit_breaker.initialize()

    def __call__(self, next: AsyncMiddleware) -> AsyncMiddleware:
        async def handle(
            req: HTTPRequest,
            client_name: ClientName,
            path: Path,
            timeout: HTTPTimeout,
        ) -> HTTPResponse:

            async with await self.circuit_breaker.get_breaker(client_name):
                resp = await next(req, client_name, path, timeout)
            return resp

        return handle



from .auth import SyncHTTPAuthorization, SyncHTTPBearerAuthorization
from .base import SyncHTTPAddHeadersMiddleware, SyncHTTPMiddleware, SyncMiddleware
from .circuit_breaker import SyncCircuitBreaker
from .http_caching import SyncHTTPCachingMiddleware
from .prometheus import SyncPrometheusMetrics
from .zipkin import SyncZipkinMiddleware

__all__ = [
    "SyncCircuitBreaker",
    "SyncHTTPAddHeadersMiddleware",
    "SyncHTTPAuthorization",
    "SyncHTTPBearerAuthorization",
    "SyncHTTPCachingMiddleware",
    "SyncHTTPMiddleware",
    "SyncMiddleware",
    "SyncPrometheusMetrics",
    "SyncZipkinMiddleware",
]



from typing import Any, Coroutine

from typing_extensions import Protocol

from blacksmith.domain.model.http import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.typing import ClientName, Path


class AsyncMiddleware(Protocol):
    """Signature of the middleware for the async version."""

    def __call__(
        self,
        req: HTTPRequest,
        client_name: ClientName,
        path: Path,
        timeout: HTTPTimeout,
    ) -> Coroutine[Any, Any, HTTPResponse]:
        """This is the next function of the middleware."""
        ...


class SyncMiddleware(Protocol):
    """Signature of the middleware for the sync version."""

    def __call__(
        self,
        req: HTTPRequest,
        client_name: ClientName,
        path: Path,
        timeout: HTTPTimeout,
    ) -> HTTPResponse:
        """This is the next function of the middleware."""
        ...



from typing import Any, Type

from blacksmith.typing import (
    ClientName,
    HTTPMethod,
    Json,
    Path,
    ResourceName,
    Service,
    ServiceName,
    Version,
)


class ConfigurationError(Exception):
    """Raised if there is a conflict for client and service."""

    def __init__(self, client: ClientName, service: Service, other: Service) -> None:
        super().__init__(
            f"Client {client} has been registered twice with "
            f"{service[0]}/{service[1]} and {other[0]}/{other[1]}"
        )


class UnregisteredServiceException(RuntimeError):
    """Raised when an unregistered service is beeing requested."""

    def __init__(self, service: ServiceName, version: Version) -> None:
        srv = f"{service}/{version}" if version else service
        super().__init__(f"Unregistered service '{srv}'")


class UnregisteredClientException(RuntimeError):
    """Raised when an unregistered service is beeing requested."""

    def __init__(self, client: ClientName) -> None:
        super().__init__(f"Unregistered client '{client}'")


class UnregisteredResourceException(RuntimeError):
    """Raised when an unregistered service is beeing requested."""

    def __init__(self, resource: ResourceName, client: ClientName) -> None:
        super().__init__(f"Unregistered resource '{resource}' in client '{client}'")


class UnregisteredRouteException(RuntimeError):
    """Raised when an unregistered service is beeing requested."""

    def __init__(
        self, route: HTTPMethod, resource: ResourceName, client: ClientName
    ) -> None:
        super().__init__(
            f"Unregistered route '{route}' in resource '{resource}' in "
            f"client '{client}'"
        )


class NoContractException(RuntimeError):
    """Raised when an unregistered service is beeing requested."""

    def __init__(
        self, method: HTTPMethod, resource: ResourceName, client: ClientName
    ) -> None:
        super().__init__(
            f"Unregistered route '{method}' in resource '{resource}' in "
            f"client '{client}'"
        )


class NoResponseSchemaException(RuntimeError):
    """Raised when an unregistered service is beeing requested."""

    def __init__(
        self, method: HTTPMethod, path: Path, resource: ResourceName, client: ClientName
    ) -> None:
        super().__init__(
            f"No response schema in route '{method} {path}' in resource'{resource}' "
            f"in client '{client}'"
        )


class WrongRequestTypeException(TypeError):
    def __init__(
        self,
        type: Type[Any],
        route: HTTPMethod,
        resource: ResourceName,
        client: ClientName,
    ) -> None:
        super().__init__(
            f"Invalid type '{type.__module__}.{type.__name__}' for route '{route}' "
            f"in resource '{resource}' in client '{client}'"
        )


class HTTPError(Exception):
    """Represent the http error."""

    from .model.http import HTTPRequest, HTTPResponse

    def __init__(self, message: str, request: HTTPRequest, response: HTTPResponse):
        super().__init__(message)
        self.request = request
        self.response = response

    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def json(self) -> Json:
        return self.response.json

    @property
    def is_client_error(self) -> bool:
        return 400 <= self.status_code < 500

    @property
    def is_server_error(self) -> bool:
        return 500 <= self.status_code < 600


class HTTPTimeoutError(TimeoutError):
    """Represent the http timeout error."""



"""
The discovery to start simple.

The static discovery strategy is a good start when you don't need a strategy.

For instance, a short list of services with static endpoint.
"""
from typing import Mapping

from blacksmith.domain.exceptions import UnregisteredServiceException
from blacksmith.typing import Service, ServiceName, Version

from ..base import SyncAbstractServiceDiscovery, Url

Endpoints = Mapping[Service, Url]


class SyncStaticDiscovery(SyncAbstractServiceDiscovery):
    """
    A discovery instance based on a static dictionary.
    """

    endpoints: Endpoints

    def __init__(self, endpoints: Endpoints) -> None:
        self.endpoints = endpoints

    def get_endpoint(self, service: ServiceName, version: Version) -> Url:
        """
        Retrieve endpoint using the given parameters from `endpoints`.
        """
        try:
            return self.endpoints[(service, version)]
        except KeyError:
            raise UnregisteredServiceException(service, version)



import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("blacksmith").version
except pkg_resources.DistributionNotFound:
    # read the doc does not support poetry
    pass

from .domain.exceptions import HTTPError, HTTPTimeoutError
from .domain.model import (
    AbstractCollectionParser,
    AbtractTraceContext,
    CollectionIterator,
    HeaderField,
    HTTPTimeout,
    PathInfoField,
    PostBodyField,
    QueryStringField,
    Request,
    Response,
    ResponseBox,
)
from .domain.registry import register
from .domain.scanner import scan
from .middleware._async import (
    AsyncCircuitBreaker,
    AsyncHTTPAddHeadersMiddleware,
    AsyncHTTPAuthorization,
    AsyncHTTPBearerAuthorization,
    AsyncHTTPCachingMiddleware,
    AsyncHTTPMiddleware,
    AsyncMiddleware,
    AsyncPrometheusMetrics,
    AsyncZipkinMiddleware,
)
from .middleware._sync import (
    SyncCircuitBreaker,
    SyncHTTPAddHeadersMiddleware,
    SyncHTTPAuthorization,
    SyncHTTPBearerAuthorization,
    SyncHTTPCachingMiddleware,
    SyncHTTPMiddleware,
    SyncMiddleware,
    SyncPrometheusMetrics,
)
from .sd._async.adapters.consul import AsyncConsulDiscovery
from .sd._async.adapters.router import AsyncRouterDiscovery
from .sd._async.adapters.static import AsyncStaticDiscovery
from .sd._sync.adapters.consul import SyncConsulDiscovery
from .sd._sync.adapters.router import SyncRouterDiscovery
from .sd._sync.adapters.static import SyncStaticDiscovery
from .service._async.client import AsyncClient, AsyncClientFactory
from .service._sync.client import SyncClient, SyncClientFactory

__all__ = [
    # Ordered for the doc
    # Request / Response
    "scan",
    "register",
    "Request",
    "Response",
    "HeaderField",
    "PathInfoField",
    "PostBodyField",
    "QueryStringField",
    # Request / Response Boxing
    "ResponseBox",
    "CollectionIterator",
    "AbstractCollectionParser",
    # Exceptions
    "HTTPError",
    "HTTPTimeoutError",
    # Timeout Config
    "HTTPTimeout",
    # Factories
    "AsyncClientFactory",
    "SyncClientFactory",
    "AsyncClient",
    "SyncClient",
    # Service Discovery
    "AsyncConsulDiscovery",
    "SyncConsulDiscovery",
    "AsyncRouterDiscovery",
    "SyncRouterDiscovery",
    "AsyncStaticDiscovery",
    "SyncStaticDiscovery",
    # Middlewares
    "AsyncMiddleware",
    "SyncMiddleware",
    # Basic Middlewares
    "AsyncHTTPMiddleware",
    "SyncHTTPMiddleware",
    "AsyncHTTPAddHeadersMiddleware",
    "SyncHTTPAddHeadersMiddleware",
    "AsyncHTTPAuthorization",
    "SyncHTTPAuthorization",
    "AsyncHTTPBearerAuthorization",
    "SyncHTTPBearerAuthorization",
    # Advanced Middlewares
    "AsyncCircuitBreaker",
    "SyncCircuitBreaker",
    "AsyncPrometheusMetrics",
    "SyncPrometheusMetrics",
    "AsyncHTTPCachingMiddleware",
    "SyncHTTPCachingMiddleware",
    "AbtractTraceContext",
    "AsyncZipkinMiddleware",
]



import abc
from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    cast,
)

from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from pydantic.typing import IntStr
else:
    IntStr = str

from ...domain.exceptions import NoResponseSchemaException
from ...typing import (
    ClientName,
    HttpLocation,
    HTTPMethod,
    Json,
    Path,
    ResourceName,
    Url,
)
from .http import HTTPRequest, HTTPResponse, Links

PATH: HttpLocation = "path"
HEADER: HttpLocation = "headers"
QUERY: HttpLocation = "querystring"
BODY: HttpLocation = "body"


PathInfoField = partial(Field, location=PATH)
"""Declare field that are serialized to the path info."""
HeaderField = partial(Field, location=HEADER)
"""Declare field that are serialized in http request header."""
QueryStringField = partial(Field, location=QUERY)
"""Declare field that are serialized in the http querystring."""
PostBodyField = partial(Field, location=BODY)
"""Declare field that are serialized in the json document."""


class Request(BaseModel):
    """
    Request Params Model.

    Fields must use subclass :func:`.PathInfoField`, :func:`.HeaderField`,
    :func:`.QueryStringField` or :func:`.PostBodyField` to declare each fields.
    """

    def to_http_request(self, method: HTTPMethod, url_pattern: Url) -> HTTPRequest:
        """Convert the request params to an http request in order to serialize
        the http request for the client.
        """
        req = HTTPRequest(method, url_pattern)
        fields_by_loc: Dict[HttpLocation, Dict[IntStr, Any]] = {
            HEADER: {},
            PATH: {},
            QUERY: {},
            BODY: {},
        }
        for field in self.__fields__.values():
            loc = cast(
                HttpLocation,
                field.field_info.extra["location"],  # type: ignore
            )
            fields_by_loc[loc].update({field.name: ...})

        headers = self.dict(
            include=fields_by_loc[HEADER], by_alias=True, exclude_none=True
        )
        req.headers = {key: str(val) for key, val in headers.items()}
        req.path = self.dict(
            include=fields_by_loc[PATH], by_alias=True, exclude_none=False
        )
        req.querystring = self.dict(
            include=fields_by_loc[QUERY], by_alias=True, exclude_none=True
        )
        req.body = self.json(
            include=fields_by_loc[BODY], by_alias=True, exclude_none=False
        )
        return req


TResponse = TypeVar("TResponse", bound="Response")
TCollectionResponse = TypeVar("TCollectionResponse", bound="Response")


class Response(BaseModel):
    """Response Model."""

    @classmethod
    def from_http_response(
        cls: Type[TResponse], response: HTTPResponse
    ) -> Optional[TResponse]:
        """Build the response from the given HTTPResponse."""
        return cls(**response.json) if response.json else None


@dataclass
class Metadata:
    """Metadata of a collection response."""

    count: int
    total_count: Optional[int]
    links: Links


class AbstractCollectionParser(abc.ABC):
    """
    Signature of the collection parser.
    """

    resp: HTTPResponse

    def __init__(self, resp: HTTPResponse):
        self.resp = resp

    @property
    @abc.abstractmethod
    def meta(self) -> Metadata:
        """
        Return the metatadata from the response.

        Usually, metadata are in a header, but if the API wrap the list,

        ::

            {
                "total_items": 0,
                "items": []
            }


        Then, the ``Metadata.total_count`` can be extracted from the json,
        instead of the header.
        """

    @property
    @abc.abstractmethod
    def json(self) -> List[Any]:
        """
        Return the list part of the response the response.

        For instance, if an API wrap the list in a structure like

        ::

            {
                "items": [
                    {"objkey": "objval"}
                ]
            }

        then, the ``resp.json["items"]`` has to be returned.
        """


class CollectionParser(AbstractCollectionParser):
    """
    Handle the rest collection metadata parser.

    Deserialize how a collection is wrapped.
    """

    total_count_header: str = "Total-Count"

    @property
    def meta(self) -> Metadata:
        total_count = self.resp.headers.get(self.total_count_header)
        return Metadata(
            count=len(self.json),
            total_count=None if total_count is None else int(total_count),
            links=self.resp.links,
        )

    @property
    def json(self) -> List[Json]:
        return self.resp.json or []


class ResponseBox(Generic[TResponse]):
    """
    Wrap an http response to deseriaze it.

    It's also allow users to write some userfull typing inference such as:

    ::

        user: User = (await api.user.get({"username": username})).response
        print(user.username)  # declaring the type User make code analyzer happy.
    """

    def __init__(
        self,
        response: HTTPResponse,
        response_schema: Optional[Type[Response]],
        method: HTTPMethod,
        path: Path,
        name: ResourceName,
        client_name: ClientName,
    ) -> None:
        self.http_response = response
        self.response_schema = response_schema
        self.method: HTTPMethod = method
        self.path: Path = path
        self.name: ResourceName = name
        self.client_name: ClientName = client_name

    @property
    def json(self) -> Optional[Dict[str, Any]]:
        """Return the raw json response."""
        return self.http_response.json

    @property
    def response(self) -> TResponse:
        """
        Parse the response using the schema.

        :raises NoResponseSchemaException: if the response_schema has not been
            set in the contract.
        """
        if self.response_schema is None:
            raise NoResponseSchemaException(
                self.method, self.path, self.name, self.client_name
            )
        resp = self.response_schema(**(self.json or {}))
        return cast(TResponse, resp)


class CollectionIterator(Iterator[TResponse]):
    """
    Deserialize the models in a json response list, item by item.
    """

    response: AbstractCollectionParser

    def __init__(
        self,
        response: HTTPResponse,
        response_schema: Optional[Type[Response]],
        collection_parser: Type[AbstractCollectionParser],
    ) -> None:
        self.pos = 0
        self.response_schema = response_schema
        self.response = collection_parser(response)
        self.json_resp = self.response.json

    @property
    def meta(self) -> Metadata:
        """
        Get the response metadata such as counts in http header, links...

        Those metadata are generated by the collection_parser.
        """
        return self.response.meta

    def __next__(self) -> TResponse:
        try:
            resp = self.json_resp[self.pos]
            if self.response_schema:
                resp = self.response_schema(**resp)
        except IndexError:
            raise StopIteration()

        self.pos += 1
        return cast(TResponse, resp)  # Could be a dict

    def __iter__(self) -> "CollectionIterator[TResponse]":
        return self



"""Collect metrics based on prometheus."""
import time
from typing import TYPE_CHECKING, Any, List, Optional

import pkg_resources

from blacksmith.domain.exceptions import HTTPError
from blacksmith.domain.model.http import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.typing import ClientName, Path

from .base import SyncHTTPMiddleware, SyncMiddleware

if TYPE_CHECKING:
    try:
        import prometheus_client  # type: ignore
    except ImportError:
        pass
    Registry = Optional["prometheus_client.CollectorRegistry"]
else:
    Registry = Any


class SyncPrometheusMetrics(SyncHTTPMiddleware):
    """
    Collect the api calls made in a prometheus registry.

    It expose a `blacksmith_info` Gauge to get the blacksmith version, as a label,
    and a `blacksmith_request_latency_seconds_count` Counter to get the number of
    http requests made.
    The counter `blacksmith_request_latency_seconds_count` as client_name, method,
    path and status_code labels.

    .. note::

        the service_name and service version is redundant with the client_name,
        so they are not exposed as labels. By the way, you may have multiple
        client_name for 1 service name/version.

    """

    def __init__(
        self, buckets: Optional[List[float]] = None, registry: Registry = None
    ) -> None:
        from prometheus_client import (  # type: ignore
            REGISTRY,
            Counter,
            Gauge,
            Histogram,
        )

        if registry is None:
            registry = REGISTRY
        if buckets is None:
            buckets = [0.05 * 2 ** x for x in range(10)]
        version_info = {"version": pkg_resources.get_distribution("blacksmith").version}
        self.blacksmith_info = Gauge(
            "blacksmith_info",
            "Blacksmith Information",
            registry=registry,
            labelnames=list(version_info.keys()),
        )
        self.blacksmith_info.labels(**version_info).set(1)

        self.blacksmith_request_latency_seconds = Histogram(
            "blacksmith_request_latency_seconds",
            "Latency of http requests in seconds",
            buckets=buckets,
            registry=registry,
            labelnames=["client_name", "method", "path", "status_code"],
        )

        self.blacksmith_circuit_breaker_error = Counter(
            "blacksmith_circuit_breaker_error",
            "Count the circuit breaker exception raised",
            registry=registry,
            labelnames=["client_name"],
        )

        self.blacksmith_circuit_breaker_state = Gauge(
            "blacksmith_circuit_breaker_state",
            "State of the circuit breaker. 0 is closed, 1 is half-opened, 2 is opened.",
            registry=registry,
            labelnames=["client_name"],
        )

    def __call__(self, next: SyncMiddleware) -> SyncMiddleware:
        def handle(
            req: HTTPRequest,
            client_name: ClientName,
            path: Path,
            timeout: HTTPTimeout,
        ) -> HTTPResponse:
            status_code = 0
            start = time.perf_counter()
            try:
                resp = next(req, client_name, path, timeout)
                status_code = resp.status_code
            except HTTPError as exc:
                status_code = exc.response.status_code
                raise exc
            finally:
                if status_code > 0:
                    latency = time.perf_counter() - start
                    metric = self.blacksmith_request_latency_seconds
                    metric.labels(
                        client_name,
                        req.method,
                        path,
                        status_code,
                    ).observe(latency)
            return resp

        return handle



from .http import HTTPRequest, HTTPResponse, HTTPTimeout
from .params import (
    AbstractCollectionParser,
    CollectionIterator,
    CollectionParser,
    HeaderField,
    PathInfoField,
    PostBodyField,
    QueryStringField,
    Request,
    Response,
    ResponseBox,
    TResponse,
)
from .tracing import AbtractTraceContext

__all__ = [
    "AbtractTraceContext",
    "AbstractCollectionParser",
    "CollectionIterator",
    "CollectionParser",
    "HeaderField",
    "HTTPRequest",
    "HTTPResponse",
    "HTTPTimeout",
    "PathInfoField",
    "PostBodyField",
    "QueryStringField",
    "Request",
    "Response",
    "ResponseBox",
    "TResponse",
]



from typing import Any, Optional, Tuple

from httpx import _types  # type: ignore
from typing_extensions import Literal

Url = str
ServiceName = str
ClientName = str
ResourceName = str
Version = Optional[str]
Service = Tuple[str, Version]

Path = str
Proxies = _types.ProxiesTypes

HttpLocation = Literal["path", "headers", "querystring", "body"]
HTTPMethod = Literal["HEAD", "GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"]

Json = Optional[Any]



from typing import Generic, List, Optional, Type

from blacksmith.domain.exceptions import UnregisteredResourceException
from blacksmith.domain.model.http import HTTPTimeout
from blacksmith.domain.model.params import (
    AbstractCollectionParser,
    CollectionParser,
    TCollectionResponse,
    TResponse,
)
from blacksmith.domain.registry import Registry, Resources
from blacksmith.domain.registry import registry as default_registry
from blacksmith.middleware._sync.base import SyncHTTPMiddleware
from blacksmith.sd._sync.base import SyncAbstractServiceDiscovery
from blacksmith.service._sync.adapters.httpx import SyncHttpxTransport
from blacksmith.typing import ClientName, Proxies, ResourceName, Url

from .base import SyncAbstractTransport
from .route_proxy import ClientTimeout, SyncRouteProxy, build_timeout


class SyncClient(Generic[TCollectionResponse, TResponse]):
    """
    Client representation for the client name.

    A client will have dymanic property, based on the registered resources.
    """

    name: ClientName
    endpoint: Url
    resources: Resources
    transport: SyncAbstractTransport
    timeout: HTTPTimeout
    collection_parser: Type[AbstractCollectionParser]
    middlewares: List[SyncHTTPMiddleware]

    def __init__(
        self,
        name: ClientName,
        endpoint: Url,
        resources: Resources,
        transport: SyncAbstractTransport,
        timeout: HTTPTimeout,
        collection_parser: Type[AbstractCollectionParser],
        middlewares: List[SyncHTTPMiddleware],
    ) -> None:
        self.name = name
        self.endpoint = endpoint
        self.resources = resources
        self.transport = transport
        self.timeout = timeout
        self.collection_parser = collection_parser
        self.middlewares = middlewares.copy()

    def add_middleware(
        self, middleware: SyncHTTPMiddleware
    ) -> "SyncClient[TCollectionResponse, TResponse]":
        self.middlewares.insert(0, middleware)
        return self

    def __getattr__(
        self, name: ResourceName
    ) -> SyncRouteProxy[TCollectionResponse, TResponse]:
        """
        The client has attributes that are the registered resource.

        The resource are registered using the :func:`blacksmith.register` function.
        """
        try:
            return SyncRouteProxy(
                self.name,
                name,
                self.endpoint,
                self.resources[name],
                self.transport,
                self.timeout,
                self.collection_parser,
                self.middlewares,
            )
        except KeyError:
            raise UnregisteredResourceException(name, self.name)


class SyncClientFactory(Generic[TCollectionResponse, TResponse]):
    """
    Client creator, for the given configuration.

    :param sd: Service Discovery instance
    :param transport: HTTP Client that process the call,
        default use :class:`blacksmith.service._async.adapters.httpx.HttpxTransport`
    :param timeout: configure timeout,
        this parameter is ignored if the transport has been passed
    :param proxies: configure proxies,
        this parameter is ignored if the transport has been passed
    :param verify_certificate: Reject request if certificate are invalid for https
    :param collection_parser: use to customize the collection parser
        default use :class:`blacksmith.domain.model.params.CollectionParser`
    """

    sd: SyncAbstractServiceDiscovery
    registry: Registry
    transport: SyncAbstractTransport
    timeout: HTTPTimeout
    collection_parser: Type[AbstractCollectionParser]
    middlewares: List[SyncHTTPMiddleware]

    def __init__(
        self,
        sd: SyncAbstractServiceDiscovery,
        transport: Optional[SyncAbstractTransport] = None,
        registry: Registry = default_registry,
        timeout: ClientTimeout = HTTPTimeout(),
        proxies: Optional[Proxies] = None,
        verify_certificate: bool = False,
        collection_parser: Type[AbstractCollectionParser] = CollectionParser,
    ) -> None:
        self.sd = sd
        self.registry = registry
        self.transport = transport or SyncHttpxTransport(
            verify_certificate=verify_certificate,
            proxies=proxies,
        )
        self.timeout = build_timeout(timeout)
        self.collection_parser = collection_parser
        self.middlewares = []
        self._initialized = False

    def add_middleware(
        self, middleware: SyncHTTPMiddleware
    ) -> "SyncClientFactory[TCollectionResponse,TResponse]":
        """
        Add a middleware to the client factory and return the client for chaining.

        ..note:: Clients created before the call of this method will also be
            altered. The middleware stack is a reference for all clients.
        """
        self.middlewares.insert(0, middleware)
        return self

    def initialize(self) -> None:
        for middleware in self.middlewares:
            middleware.initialize()

    def __call__(
        self, client_name: ClientName
    ) -> SyncClient[TCollectionResponse, TResponse]:
        if not self._initialized:
            self._initialized = True
            self.initialize()
        srv, resources = self.registry.get_service(client_name)
        endpoint = self.sd.get_endpoint(srv[0], srv[1])
        return SyncClient(
            client_name,
            endpoint,
            resources,
            self.transport,
            self.timeout,
            self.collection_parser,
            self.middlewares,
        )



"""
Convenient method to fillout an blacksmith registry.

::

  import blacksmith
  blacksmith.scan('mypkg.resources', 'other.resources')
"""
import importlib
import pkgutil


def scan(*modules: str) -> None:
    """
    Collect all resources to fillout the registry.

    Basically, it import modules registered using :func:`blacksmith.register`.

    :raises TypeError: malformed module name
    :raises ModuleNotFoundError: unknown package name
    :raises AttributeError: argument is a module, not a package.
    """
    for modname in modules:
        if modname.startswith("."):
            raise ValueError(f"{modname}: Relative package unsupported")
        mod = importlib.import_module(modname)
        if hasattr(mod, "__path__"):  # it means it is a __init__.py.
            for _loader, submod, _is_pkg in pkgutil.walk_packages(
                path=mod.__path__,
                prefix=mod.__name__ + ".",
            ):
                importlib.import_module(submod)



from typing import Mapping

import httpx
from httpx import Response as HttpxRepsonse
from httpx import Timeout as HttpxTimeout

from blacksmith.domain.exceptions import HTTPError, HTTPTimeoutError
from blacksmith.domain.model import HTTPRequest, HTTPResponse, HTTPTimeout
from blacksmith.service.ports import AsyncClient
from blacksmith.typing import ClientName, Json, Path

from ..base import AsyncAbstractTransport


def safe_json(r: HttpxRepsonse) -> Json:
    try:
        return r.json()
    except Exception:
        return {"error": r.text}


class AsyncHttpxTransport(AsyncAbstractTransport):
    """
    Transport implemented using `httpx`_.

    .. _`httpx`: https://www.python-httpx.org/

    """

    async def __call__(
        self,
        req: HTTPRequest,
        client_name: ClientName,
        path: Path,
        timeout: HTTPTimeout,
    ) -> HTTPResponse:
        headers = req.headers.copy()
        if req.body:
            headers["Content-Type"] = "application/json"
        async with AsyncClient(
            verify=self.verify_certificate,
            proxies=self.proxies,  # type: ignore
        ) as client:
            try:
                r = await client.request(  # type: ignore
                    req.method,
                    req.url,
                    params=req.querystring,
                    headers=headers,
                    content=req.body,
                    timeout=HttpxTimeout(timeout.request, connect=timeout.connect),
                )
            except httpx.TimeoutException as exc:
                raise HTTPTimeoutError(
                    f"{client_name} - {req.method} {path} - "
                    f"{exc.__class__.__name__} while calling {req.method} {req.url}"
                )

        status_code: int = r.status_code  # type: ignore
        headers: Mapping[str, str] = r.headers  # type: ignore
        json = "" if status_code == 204 else safe_json(r)
        resp = HTTPResponse(status_code, headers, json=json)
        if not r.is_success:
            raise HTTPError(
                f"{client_name} - {req.method} {path} - "
                f"{r.status_code} {r.reason_phrase}",
                req,
                resp,
            )
        return resp



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['blacksmith',
 'blacksmith.domain',
 'blacksmith.domain.model',
 'blacksmith.middleware',
 'blacksmith.middleware._async',
 'blacksmith.middleware._sync',
 'blacksmith.sd',
 'blacksmith.sd._async',
 'blacksmith.sd._async.adapters',
 'blacksmith.sd._sync',
 'blacksmith.sd._sync.adapters',
 'blacksmith.service',
 'blacksmith.service._async',
 'blacksmith.service._async.adapters',
 'blacksmith.service._sync',
 'blacksmith.service._sync.adapters']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.21.1,<0.22.0',
 'purgatory-circuitbreaker>=0.7.2,<0.8.0',
 'pydantic>=1.8.2,<2.0.0',
 'typing-extensions>=4.0.1,<5.0.0']

extras_require = \
{'caching': ['aioredis>=2.0.0,<3.0.0'],
 'prometheus': ['prometheus-client>=0.12.0,<0.13.0']}

setup_kwargs = {
    'name': 'blacksmith',
    'version': '0.12.1',
    'description': 'REST API Client designed for microservices',
    'long_description': 'Blacksmith\n==========\n\n.. image:: https://readthedocs.org/projects/python-blacksmith/badge/?version=latest\n   :target: https://python-blacksmith.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://github.com/mardiros/blacksmith/actions/workflows/main.yml/badge.svg\n   :target: https://github.com/mardiros/blacksmith/actions/workflows/main.yml\n   :alt: Continuous Integration\n\n.. image:: https://codecov.io/gh/mardiros/blacksmith/branch/master/graph/badge.svg?token=17KAC0LW9H\n   :target: https://codecov.io/gh/mardiros/blacksmith\n   :alt: Coverage Report\n\n\nBlacksmith is a library to build a solid microservices architecture based on REST API.\n\nTodays, developers have lots of choices to create microservices,\nplenty of framework are available, but when it comes to consume them,\nthere is a lack of clients.\n\nConsuming an API is not just about doing HTTP requests, it has to be designed\nfor failure, monitoring, and service discovery with an elegant abstraction.\nblacksmith aims to provide a solution for developers to write clean client code\nand for ops to monitor api calls also on the client side.\n\n\nWhat is Blacksmith\n------------------\n\nBlacksmith is a declarative tool for consuming REST Api.\n\nIn a REST API, resources are declared under HTTP routes, and every http verb\nhas its own definition.\n\nIn Blacksmith, every resources are bound to schemas that define request and response,\nin order to abstract HTTP.\n\nThis is a common concept for SQL table with ORM, where tables are bound to models,\nand then, operations are available on models. This is a usefull abstraction to \nwrite maintainable code and to dive into a project easilly.\n\nHandling API resources using an http client, such as `requests`_ does not handle\nthat abstraction, and does not handle bindings to objects, and can be compared to\na raw connection because it is just a transport.\n\nThis is the problem blacksmith is trying to solve, having a nice abstraction of\nservices that use REST in a microservices architecture.\n\n.. note::\n\n   | Blacksmith is not an HTTP Client or a model validator.\n   | Blacksmith use `httpx`_ to perform http query, and use `Pydantic`_ to validate models.\n\n.. _`requests`: https://docs.python-requests.org/\n.. _`httpx`: https://www.python-httpx.org/\n.. _`Pydantic`: https://pydantic-docs.helpmanual.io/\n\n\nWhy not using a SDK to consume APIs ?\n-------------------------------------\n\nSDK are about importing an external library in a service. And a service is\nconsumed by many services for different purpose. As a result, SDK create\ncoupling between service, and this is something that should be avoid.\n\nAn SDK for a service will declare all the resources, routes, and attribute\nof resources when a service consumer may consume just a few.\n\nSDK may hide what is really used by every service.\n\nTo avoid this, every consumers of API, should declare its own consumers\ncontracts to get a better view of which service use what.\n\n.. note::\n\n   TLDR; SDK are fine in public API, by the way, but not in a microservices\n   architecture.\n\n\nBuilding SDK\n------------\n\nBy the way, public API provider comes with an SDK, which is a good case,\nand blacksmith can be used to build SDK for Python / asyncio. \n\n\nWorking with or without asyncio\n-------------------------------\n\nBlacksmith is written for ``asyncio``, but works with synchronous API without overhead.\nThe synchronous API is generated using `unasync`_ and tested.\n\n.. note::\n\n   All class started by `Async` are asynchronous, and all class started by `Sync` are\n   synchronous, for an explicit usage\n\n.. _`unasync`: https://unasync.readthedocs.io/en/latest/\n\n\nRead More\n---------\n\nYou can read the `full documentation of this library here`_.\n\n.. _`full documentation of this library here`: https://python-blacksmith.readthedocs.io/en/latest/user/introduction.html\n',
    'author': 'Guillaume Gauvrit',
    'author_email': 'guillaume@gauvr.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mardiros/blacksmith',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
The Server-Side Service discovery strategy.

"""
from blacksmith.typing import ServiceName, Version

from ..base import AsyncAbstractServiceDiscovery, Url


class AsyncRouterDiscovery(AsyncAbstractServiceDiscovery):
    """
    Router that implement a :term:`Server-Side Service discovery`.

    This implementation never raise
    :class:`blacksmith.domain.exceptions.UnregisteredServiceException`
    when service are unknown, because it only passe very request to the router
    server that is connected to the :term:`service registry`.

    .. note::
        Given pattern in parameter have to match the format of the router server.

    :param service_url_fmt: A pattern used to create endpoint of versionned services.
    :param unversioned_service_url_fmt: A pattern used to create endpoint of
        unversionned services.
    """

    service_url_fmt: str
    unversioned_service_url_fmt: str

    def __init__(
        self,
        service_url_fmt: str = "http://router/{service}-{version}/{version}",
        unversioned_service_url_fmt: str = "http://router/{service}",
    ) -> None:
        self.service_url_fmt = service_url_fmt
        self.unversioned_service_url_fmt = unversioned_service_url_fmt

    async def get_endpoint(self, service: ServiceName, version: Version) -> Url:
        """
        Create and return the endpoint using the given parameters `service_url_fmt`
        or `unversioned_service_url_fmt` if version is `None`.
        """
        if version is None:
            name = self.unversioned_service_url_fmt.format(service=service)
        else:
            name = self.service_url_fmt.format(service=service, version=version)
        return name



from typing import Any, Dict, Generic, List, Optional, Tuple, Type, Union

from blacksmith.domain.exceptions import (
    NoContractException,
    UnregisteredRouteException,
    WrongRequestTypeException,
)
from blacksmith.domain.model import (
    CollectionIterator,
    HTTPRequest,
    HTTPResponse,
    HTTPTimeout,
    Request,
    Response,
    ResponseBox,
)
from blacksmith.domain.model.params import (
    AbstractCollectionParser,
    TCollectionResponse,
    TResponse,
)
from blacksmith.domain.registry import ApiRoutes, HttpCollection, HttpResource
from blacksmith.domain.typing import SyncMiddleware
from blacksmith.middleware._sync.base import SyncHTTPMiddleware
from blacksmith.typing import ClientName, HTTPMethod, Path, ResourceName, Url

from .base import SyncAbstractTransport

ClientTimeout = Union[HTTPTimeout, float, Tuple[float, float]]
HTTPAuthentication = SyncHTTPMiddleware


def build_timeout(timeout: ClientTimeout) -> HTTPTimeout:
    """Build the timeout from the convenient timeout."""
    if isinstance(timeout, float):
        timeout = HTTPTimeout(timeout)
    elif isinstance(timeout, tuple):
        timeout = HTTPTimeout(*timeout)
    return timeout


class SyncRouteProxy(Generic[TCollectionResponse, TResponse]):
    """Proxy from resource to its associate routes."""

    client_name: ClientName
    name: ResourceName
    endpoint: Url
    routes: ApiRoutes
    transport: SyncAbstractTransport
    timeout: HTTPTimeout
    collection_parser: Type[AbstractCollectionParser]
    middlewares: List[SyncHTTPMiddleware]

    def __init__(
        self,
        client_name: ClientName,
        name: ResourceName,
        endpoint: Url,
        routes: ApiRoutes,
        transport: SyncAbstractTransport,
        timeout: HTTPTimeout,
        collection_parser: Type[AbstractCollectionParser],
        middlewares: List[SyncHTTPMiddleware],
    ) -> None:
        self.client_name = client_name
        self.name = name
        self.endpoint = endpoint
        self.routes = routes
        self.transport = transport
        self.timeout = timeout
        self.collection_parser = collection_parser
        self.middlewares = middlewares

    def _prepare_request(
        self,
        method: HTTPMethod,
        params: Union[Optional[Request], Dict[Any, Any]],
        resource: Optional[HttpResource],
    ) -> Tuple[Path, HTTPRequest, Optional[Type[Response]]]:
        if resource is None:
            raise UnregisteredRouteException(method, self.name, self.client_name)
        if resource.contract is None or method not in resource.contract:
            raise NoContractException(method, self.name, self.client_name)

        param_schema, return_schema = resource.contract[method]
        if isinstance(params, dict):
            params = param_schema(**params)
        elif params is None:
            params = param_schema()
        elif not isinstance(params, param_schema):
            raise WrongRequestTypeException(
                params.__class__,  # type: ignore
                method,
                self.name,
                self.client_name,
            )
        req = params.to_http_request(method, self.endpoint + resource.path)
        return (resource.path, req, return_schema)

    def _prepare_response(
        self,
        response: HTTPResponse,
        response_schema: Optional[Type[Response]],
        method: HTTPMethod,
        path: Path,
    ) -> ResponseBox[TResponse]:
        return ResponseBox[TResponse](
            response,
            response_schema,
            method,
            path,
            self.name,
            self.client_name,
        )

    def _prepare_collection_response(
        self,
        response: HTTPResponse,
        response_schema: Optional[Type[Response]],
        collection_parser: Optional[Type[AbstractCollectionParser]],
    ) -> CollectionIterator[TCollectionResponse]:

        return CollectionIterator(
            response, response_schema, collection_parser or self.collection_parser
        )

    def _handle_req_with_middlewares(
        self, req: HTTPRequest, timeout: HTTPTimeout, path: Path
    ) -> HTTPResponse:
        next: SyncMiddleware = self.transport
        for middleware in self.middlewares:
            next = middleware(next)

        resp = next(req, self.client_name, path, timeout)
        return resp

    def _yield_collection_request(
        self,
        method: HTTPMethod,
        params: Union[Optional[Request], Dict[Any, Any]],
        timeout: HTTPTimeout,
        collection: HttpCollection,
    ) -> CollectionIterator[TCollectionResponse]:
        path, req, resp_schema = self._prepare_request(method, params, collection)
        resp = self._handle_req_with_middlewares(req, timeout, path)
        return self._prepare_collection_response(
            resp, resp_schema, collection.collection_parser
        )

    def _collection_request(
        self,
        method: HTTPMethod,
        params: Union[Request, Dict[Any, Any]],
        timeout: HTTPTimeout,
    ) -> ResponseBox[TResponse]:
        path, req, resp_schema = self._prepare_request(
            method, params, self.routes.collection
        )
        resp = self._handle_req_with_middlewares(req, timeout, path)
        return self._prepare_response(resp, resp_schema, method, path)

    def _request(
        self,
        method: HTTPMethod,
        params: Union[Request, Dict[Any, Any]],
        timeout: HTTPTimeout,
    ) -> ResponseBox[TResponse]:
        path, req, resp_schema = self._prepare_request(
            method, params, self.routes.resource
        )
        resp = self._handle_req_with_middlewares(req, timeout, path)
        return self._prepare_response(resp, resp_schema, method, path)

    def collection_head(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._collection_request(
            "HEAD", params, build_timeout(timeout or self.timeout)
        )

    def collection_get(
        self,
        params: Union[Optional[Request], Dict[Any, Any]] = None,
        timeout: Optional[ClientTimeout] = None,
    ) -> CollectionIterator[TCollectionResponse]:
        if not self.routes.collection:
            raise UnregisteredRouteException("GET", self.name, self.client_name)
        return self._yield_collection_request(
            "GET",
            params,
            build_timeout(timeout or self.timeout),
            self.routes.collection,
        )

    def collection_post(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._collection_request(
            "POST", params, build_timeout(timeout or self.timeout)
        )

    def collection_put(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._collection_request(
            "PUT", params, build_timeout(timeout or self.timeout)
        )

    def collection_patch(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._collection_request(
            "PATCH", params, build_timeout(timeout or self.timeout)
        )

    def collection_delete(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._collection_request(
            "DELETE", params, build_timeout(timeout or self.timeout)
        )

    def collection_options(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._collection_request(
            "OPTIONS", params, build_timeout(timeout or self.timeout)
        )

    def head(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._request("HEAD", params, build_timeout(timeout or self.timeout))

    def get(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        resp = self._request("GET", params, build_timeout(timeout or self.timeout))
        return resp

    def post(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._request("POST", params, build_timeout(timeout or self.timeout))

    def put(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._request("PUT", params, build_timeout(timeout or self.timeout))

    def patch(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._request("PATCH", params, build_timeout(timeout or self.timeout))

    def delete(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._request("DELETE", params, build_timeout(timeout or self.timeout))

    def options(
        self,
        params: Union[Request, Dict[Any, Any]],
        timeout: Optional[ClientTimeout] = None,
    ) -> ResponseBox[TResponse]:
        return self._request("OPTIONS", params, build_timeout(timeout or self.timeout))



"""Authentication Middlewares."""

from .base import SyncHTTPAddHeadersMiddleware


class SyncHTTPAuthorization(SyncHTTPAddHeadersMiddleware):
    """
    Authentication Mechanism based on the header `Authorization`.

    :param scheme: the scheme of the mechanism.
    :param value: the value that authenticate the user using the scheme.
    """

    def __init__(self, scheme: str, value: str):
        return super().__init__({"Authorization": f"{scheme} {value}"})


class SyncHTTPBearerAuthorization(SyncHTTPAuthorization):
    """
    Authentication Mechanism based on the header `Authorization` with the Bearer scheme.

    :param value: value of the bearer token.
    """

    def __init__(self, bearer_token: str):
        return super().__init__("Bearer", bearer_token)
